import pytest

from arcGisFeatureCache.utils.immutableList import ImmutableList

# Test data
data = [1, 2, 3, 4, 5]
immutable_list = ImmutableList(data)


# Test cases
def test_getitem():
    assert immutable_list[0] == 1
    assert immutable_list[2] == 3
    with pytest.raises(IndexError):
        immutable_list[10]


def test_len():
    assert len(immutable_list) == len(data)


def test_contains():
    assert 3 in immutable_list
    assert 10 not in immutable_list


def test_repr():
    assert repr(immutable_list) == f"ImmutableList({tuple(data)})"


def test_index():
    assert immutable_list.index(3) == 2
    with pytest.raises(ValueError):
        immutable_list.index(10)


def test_count():
    assert immutable_list.count(3) == 1
    assert immutable_list.count(10) == 0


def test_equality():
    assert immutable_list == ImmutableList(data)
    assert immutable_list != ImmutableList([5, 4, 3, 2, 1])


def test_equality_same_lists():
    assert immutable_list == ImmutableList(data)


def test_equality_different_lists():
    assert not (immutable_list == ImmutableList([5, 4, 3, 2, 1]))


def test_equality_with_non_immutable_list():
    assert immutable_list.__eq__(1) is NotImplemented


def test_addition():
    new_list = immutable_list + ImmutableList([6, 7])
    assert len(new_list) == len(data) + 2


def test_multiplication():
    new_list = immutable_list * 3
    assert len(new_list) == len(data) * 3


def test_reflected_addition():
    new_list = [6, 7] + immutable_list
    assert len(new_list) == len(data) + 2


def test_invalid_operations():
    with pytest.raises(TypeError):
        immutable_list + [6, 7]
    with pytest.raises(TypeError):
        immutable_list * "invalid"


def test_radd_invalid_operand():
    with pytest.raises(TypeError):
        immutable_list.__radd__(1)


def test_inequality_same_lists():
    assert not (immutable_list != ImmutableList(data))


def test_inequality_different_lists():
    assert immutable_list != ImmutableList([5, 4, 3, 2, 1])


def test_inequality_with_non_immutable_list():
    assert immutable_list.__ne__(1) is NotImplemented


def test_rmul():
    n = 3
    multiplied_list = immutable_list.__rmul__(n)
    assert multiplied_list == immutable_list * n


def test_iteration():
    # Convert ImmutableList to a list for direct comparison
    assert list(immutable_list) == data



import asyncio
import json
import math
import ssl

import httpx
from httpx import Limits

from arcGisFeatureCache.utils.log import logger

# seems arcgis feature service usages LEGACY_RENEGOTIATION, its a thing on 3.11 and above
CUSTOM_SSL_CONTEXT = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
CUSTOM_SSL_CONTEXT.options |= (
    0x00040000  # OP flag SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION
)
CUSTOM_SSL_CONTEXT.check_hostname = False
CUSTOM_SSL_CONTEXT.verify_mode = ssl.CERT_NONE


class ArcGisScraper:
    """A class for scraping data from ArcGIS Feature Services.

    Attributes:
        feature_service_info: Information about the feature service.
        feature_layers: List of feature layers.
        layer_info: Information about individual layers.

    """

    def __init__(
        self,
        feature_service_url: str,
        _max_connections: int = 5,
        _max_keepalive_connections: int = 5,
    ):
        self._base_url = feature_service_url
        self.feature_service_info: dict | None = None
        self.feature_layers: list[dict] | None = None
        self._limits = Limits(
            max_connections=_max_connections,
            max_keepalive_connections=_max_keepalive_connections,
        )
        self._transport = httpx.HTTPTransport(retries=1)
        self.layer_info: dict = {}

    @classmethod
    async def factory(cls, url):
        """
        Create an ArcGisScraper instance.

        Parameters:
            url: The URL of the ArcGIS Feature Service.

        Returns:
            An instance of ArcGisScraper.

        Raises:
            ValueError: If the URL is invalid.
        """
        self = ArcGisScraper(url, 100, 20)
        await self._process_all_layers()
        return self

    async def _fetch(self, url: str) -> dict:
        """
        Asynchronously fetch data from the specified URL.

        Parameters:
            url: The URL to fetch data from.

        Returns:
            The fetched data.

        Raises:
            ValueError: If the response cannot be parsed as JSON.
        """
        async with httpx.AsyncClient(
            timeout=None, limits=self._limits, verify=CUSTOM_SSL_CONTEXT
        ) as client:
            success = False
            while not success:
                try:
                    r = await client.get(url)
                    logger.info(f"fetched data, status {r.status_code}: {url}")
                    response = json.loads(r.text)
                    if isinstance(response, dict):
                        return response
                    raise ValueError(f"Can bake json from response {url}")  # noqa: TRY003, TRY301
                except Exception as e:
                    logger.debug(f"error: {e}")
                    await asyncio.sleep(3)
        raise ValueError(f"cant fetch data {url}")  # noqa: TRY003

    async def _get_json_data_async(self, url: str) -> dict:
        """
        Asynchronously fetch JSON data from the specified URL.

        Parameters:
            url: The URL to fetch JSON data from.

        Returns:
            The fetched JSON data.
        """
        retry_seconds = 60
        json_data = await self._fetch(url)
        if "error" in json_data.keys():
            success = False
            while not success:
                logger.error(f"fetched data got error: {url}")
                logger.info(f"retry in {retry_seconds} seconds")
                await asyncio.sleep(retry_seconds)
                json_data = await self._fetch(url)
                if "error" not in json_data.keys():
                    success = True

        return json_data

    async def _get_all_data_async(
        self, base_url: str, feature_offset: int = 2000
    ) -> dict:
        """
        Asynchronously fetch all data from the feature service.

        Parameters:
            base_url: The base URL of the feature service.
            feature_offset: Offset for fetching features.

        Returns:
            All fetched data.
        """
        self.layer_info = await self._get_json_data_async(base_url + "?f=pjson")

        # get all data
        feature_service_url = (
            base_url + "/query?returnZ=true&returnM=true&f=json&outfields=*&where=1%3D1"
        )
        count_response = await self._get_json_data_async(
            feature_service_url + "&returnCountOnly=true"
        )

        # create batches
        feature_total_count = count_response["count"]
        if feature_total_count == 0:
            return {
                "dataset": base_url,
                "features": [],
                "geometryType": self.layer_info["geometryType"],
                "fields": self.layer_info["fields"],
                "objectIdFieldName": self.layer_info["objectIdField"],
                "globalIdFieldName": self.layer_info["globalIdField"],
                "spatialReference": {"latestWkid": None, "wkid": None},
            }

        batch_count = math.ceil(feature_total_count / feature_offset)
        batch_list = []
        offset = 0
        for n in range(batch_count):
            batch_list.append(feature_service_url + f"&resultOffset={offset}")
            offset += feature_offset

        # create and await all task
        tasks = [self._get_json_data_async(url) for url in batch_list]
        feature_data = await asyncio.gather(*tasks)

        logger.debug(f"Merging data {base_url}")

        features: list[dict] = sum([item["features"] for item in feature_data], [])

        feature_data[0]["features"] = features
        logger.debug(f"Done merging data {base_url}")

        # TODO: make succeeded check on feature count
        feature_data[0]["dataset"] = base_url
        if isinstance(feature_data[0], dict):
            return feature_data[0]
        else:
            raise ValueError("")  # noqa: TRY004

    async def _get_layer_data(self, feature_service_layer_url: str) -> dict:
        """
        Asynchronously fetch data for a specific layer.

        Parameters:
            feature_service_layer_url: The URL of the layer.

        Returns:
            Data for the layer.
        """
        return await self._get_all_data_async(feature_service_layer_url)

    async def _process_all_layers(self):
        """
        Asynchronously process all layers of the feature service.
        """
        self.feature_service_info = await self._get_json_data_async(
            self._base_url + "?f=pjson"
        )
        batch_list = [
            f"{self._base_url}/{item['id']}"
            for item in self.feature_service_info["layers"]
        ]
        tasks = [self._get_layer_data(url) for url in batch_list]
        feature_data = await asyncio.gather(*tasks)
        self.feature_layers = feature_data



import pytest
from shapely.geometry import LineString, Point, Polygon

from arcGisFeatureCache.models.geometry import (
    FeatureSpatialReference,
    parse_line,
    parse_point,
    parse_polygon,
)


@pytest.mark.asyncio
async def test_parse_point():
    geometry_value = {"x": 1, "y": 2}
    expected_point = Point(1, 2)
    assert await parse_point(geometry_value) == expected_point


@pytest.mark.asyncio
async def test_parse_point_empty_geometry():
    # Testing empty geometry
    assert await parse_point("") == Point()


@pytest.mark.asyncio
async def test_parse_line():
    geometry_value = {"paths": [[[1, 2], [3, 4]]]}
    expected_line = LineString([(1, 2), (3, 4)])
    assert await parse_line(geometry_value) == expected_line


@pytest.mark.asyncio
async def test_parse_line_empty_geometry():
    # Testing empty geometry
    assert await parse_line({}) == LineString()

    # Testing empty paths
    geometry_value_empty_paths = {"paths": []}
    assert await parse_line(geometry_value_empty_paths) == LineString()


@pytest.mark.asyncio
async def test_parse_line_has_z():
    geometry_value_with_z = {"paths": [[[1, 2, 3], [4, 5, 6]]], "hasZ": True}
    expected_line_with_z = LineString([(1, 2, 3), (4, 5, 6)])
    assert await parse_line(geometry_value_with_z) == expected_line_with_z


@pytest.mark.asyncio
async def test_parse_polygon():
    geometry_value = {"rings": [[[1, 2], [3, 4], [5, 6], [1, 2]]]}
    expected_polygon = Polygon([(1, 2), (3, 4), (5, 6), (1, 2)])
    assert await parse_polygon(geometry_value) == expected_polygon


@pytest.mark.asyncio
async def test_parse_polygon_empty_geometry():
    # Testing empty geometry
    assert await parse_polygon({}) == Polygon()

    # Testing empty rings
    geometry_value_empty_rings = {"rings": [[]]}
    assert await parse_polygon(geometry_value_empty_rings) == Polygon()


def test_feature_spatial_reference():
    data = {"wkid": "4326", "latestWkid": "4326"}
    spatial_ref = FeatureSpatialReference(data)
    assert spatial_ref.wkid == "4326"
    assert spatial_ref.latest_wkid == "4326"



from datetime import datetime

import pytest

from arcGisFeatureCache.models.attributes import (
    EsriFieldTypeEnum,
    FeatureAttribute,
    FeatureAttributes,
    FeatureField,
    FeatureFields,
    parse_date,
)


@pytest.fixture
def sample_feature_field_data():
    return {
        "name": "field_name",
        "alias": "Field Alias",
        "type": EsriFieldTypeEnum.esriFieldTypeString.value,
        "length": 50,
    }


@pytest.fixture
def sample_feature_attribute_data():
    return {"attr1": "value1", "attr2": "value2", "attr3": "value3"}


@pytest.fixture
def sample_feature_fields(sample_feature_field_data):
    return FeatureFields([sample_feature_field_data])


@pytest.fixture
def sample_feature_attributes(sample_feature_attribute_data):
    return FeatureAttributes(sample_feature_attribute_data)


def test_feature_field_init(sample_feature_field_data):
    field = FeatureField(sample_feature_field_data)
    assert field.name == sample_feature_field_data["name"]
    assert field.alias == sample_feature_field_data["alias"]
    assert field.type == EsriFieldTypeEnum.esriFieldTypeString
    assert field.length == sample_feature_field_data["length"]


def test_feature_fields_get_date_fields(sample_feature_fields):
    date_field = {
        "name": "date_field",
        "alias": "Date Field",
        "type": EsriFieldTypeEnum.esriFieldTypeDate.value,
    }
    sample_feature_fields._items.append(FeatureField(date_field))
    date_fields = sample_feature_fields.get_date_fields()
    assert len(date_fields) == 1
    assert date_fields[0].name == "date_field"
    assert date_fields[0].type == EsriFieldTypeEnum.esriFieldTypeDate


def test_feature_attribute_set_attribute(sample_feature_attributes):
    field = FeatureField(
        {
            "name": "attr1",
            "type": EsriFieldTypeEnum.esriFieldTypeString.value,
            "alias": "",
        }
    )
    sample_feature_attributes.set_attribute(field, lambda x: x.upper())
    assert sample_feature_attributes[0].value == "VALUE1"


def test_feature_attributes_get_all_fields(sample_feature_attributes):
    fields = sample_feature_attributes.get_all_fields()
    assert len(fields) == 3
    assert "attr1" in fields
    assert "attr2" in fields
    assert "attr3" in fields


def test_parse_date():
    timestamp = 1616198400000  # March 20, 2021
    expected_date = datetime(2021, 3, 20)
    parsed_date = parse_date(timestamp)
    assert parsed_date == expected_date


def test_feature_attribute_get_value(sample_feature_attributes):
    value = sample_feature_attributes.get_value("attr1")
    assert value == "value1"


def test_feature_attribute_get_value_nonexistent_key(sample_feature_attributes):
    value = sample_feature_attributes.get_value("nonexistent_key")
    assert value is None


def test_feature_attribute_get_value_duplicate_key(sample_feature_attributes):
    sample_feature_attributes._items.append(FeatureAttribute("attr1", "value2"))
    with pytest.raises(Exception):
        sample_feature_attributes.get_value("attr1")


def test_feature_attribute_len(sample_feature_attributes):
    assert len(sample_feature_attributes) == 3


def test_feature_attribute_delitem_not_allowed(sample_feature_attributes):
    with pytest.raises(NotImplementedError):
        del sample_feature_attributes[0]


def test_feature_attribute_setitem_not_allowed(sample_feature_attributes):
    with pytest.raises(NotImplementedError):
        sample_feature_attributes[0] = FeatureAttribute("attr", "value")


def test_feature_attribute_repr(sample_feature_attributes):
    expected_repr = "[attr1=value1, attr2=value2, attr3=value3]"
    assert repr(sample_feature_attributes) == expected_repr


def test_feature_fields_get_item_by_index(sample_feature_fields):
    field = sample_feature_fields[0]
    assert field.name == "field_name"
    assert field.alias == "Field Alias"
    assert field.type == EsriFieldTypeEnum.esriFieldTypeString
    assert field.length == 50


def test_feature_fields_get_item_by_slice(sample_feature_fields):
    fields = sample_feature_fields[:]
    assert len(fields) == 1
    assert fields[0].name == "field_name"
    assert fields[0].alias == "Field Alias"
    assert fields[0].type == EsriFieldTypeEnum.esriFieldTypeString
    assert fields[0].length == 50


def test_feature_fields_iter(sample_feature_fields):
    fields = [field for field in sample_feature_fields]
    assert len(fields) == 1
    assert fields[0].name == "field_name"
    assert fields[0].alias == "Field Alias"
    assert fields[0].type == EsriFieldTypeEnum.esriFieldTypeString
    assert fields[0].length == 50



import asyncio

from arcGisFeatureCache import ArcGisFeatureService


async def main():
    url = "https://maps.prorail.nl/arcgis/rest/services/Tekeningen_schematisch/FeatureServer"
    feature_service = await ArcGisFeatureService.factory(url)

    # get all features from service
    all_features = feature_service.get_all_features()
    first_feature = all_features[0]

    # get the layer name
    print(first_feature.dataset)

    # we use shapley for geometry 😊
    print(first_feature.geometry)

    # dynamic attributes
    print(first_feature.attributes)

    # get them by field
    fields = first_feature.attributes.get_all_fields()
    for field in fields:
        print(f"{field}: {first_feature.attributes.get_value(field)}")

    # get features from one or more layers
    layer_features = feature_service.get_layer_features(["OBE blad"])
    print(layer_features[-1].dataset)


if __name__ == "__main__":
    asyncio.run(main())



from collections.abc import Callable, Iterator
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Any, SupportsIndex, overload

from arcGisFeatureCache.utils.helpers import clear_temp_data
from arcGisFeatureCache.utils.immutableList import ImmutableList


class EsriFieldTypeEnum(Enum):
    """An enumeration representing Esri field types.

    Attributes:
        esriFieldTypeOID: esriFieldTypeOID
        esriFieldTypeString: esriFieldTypeString
        esriFieldTypeInteger: esriFieldTypeInteger
        esriFieldTypeSmallInteger: esriFieldTypeSmallInteger
        esriFieldTypeDouble: esriFieldTypeDouble
        esriFieldTypeDate: esriFieldTypeDate
        esriFieldTypeGeometry: esriFieldTypeGeometry
        esriFieldTypeBlob: esriFieldTypeBlob
    """

    esriFieldTypeOID = "esriFieldTypeOID"
    esriFieldTypeString = "esriFieldTypeString"
    esriFieldTypeInteger = "esriFieldTypeInteger"
    esriFieldTypeSmallInteger = "esriFieldTypeSmallInteger"
    esriFieldTypeDouble = "esriFieldTypeDouble"
    esriFieldTypeDate = "esriFieldTypeDate"
    esriFieldTypeGeometry = "esriFieldTypeGeometry"
    esriFieldTypeBlob = "esriFieldTypeBlob"


@dataclass
class FeatureField:
    """A class representing a field in a feature."

    Parameters:
        name (str): Name of the field.
        alias:(str): Alias name of the field.
        type :(EsriFieldTypeEnum)
        length: Length of the Field.
    """

    _data: dict[str, Any] | None

    name: str = field(init=False)
    alias: str = field(init=False)
    type: EsriFieldTypeEnum = field(init=False)
    length: int | None = field(init=False)

    def __post_init__(self) -> None:
        if self._data is not None:
            if self._data["name"] is not None:
                self.name = self._data["name"]
            self.name = self._data["name"]
            self.alias = self._data["alias"]
            self.type = EsriFieldTypeEnum[self._data["type"]]
            if "length" in self._data.keys():
                self.length = self._data["length"]
            else:
                self.length = None

        clear_temp_data(self)


@dataclass
class FeatureFields(ImmutableList):
    """A class representing a list of feature fields.

    Converts input data to a list of FeatureField instances.
    """

    _data: list | None
    _items: list[FeatureField] = field(init=False)

    def __post_init__(self) -> None:
        if self._data is not None and isinstance(self._data, list):
            self._items = [FeatureField(item) for item in self._data]
        else:
            self._items = []

        clear_temp_data(self)

    def __iter__(self) -> Iterator[FeatureField]:
        yield from self._items

    @overload
    def __getitem__(self, idx: SupportsIndex) -> FeatureField: ...

    @overload
    def __getitem__(self, idx: slice) -> list[FeatureField]: ...

    def __getitem__(
        self, idx: SupportsIndex | slice
    ) -> FeatureField | list[FeatureField]:
        """Get a feature field by index or slice."""
        return self._items[idx]

    def get_date_fields(self) -> list[FeatureField]:
        """
        Get fields of type 'date'.

        Returns:
            List[FeatureField]: List of fields of type 'esriFieldTypeDate'.
        """
        return [
            item
            for item in self._items
            if item.type == EsriFieldTypeEnum.esriFieldTypeDate
        ]


@dataclass
class FeatureAttribute:
    """A class representing an attribute of a feature.

    Parameters:
        key (str): Name of the attribute.
        value (str): Value of the attribute.
    """

    key: str
    value: Any

    def __repr__(self) -> str:
        return f"{self.key}={self.value}"


@dataclass
class FeatureAttributes(list[FeatureAttribute]):
    """A class representing a list of feature attributes.

    Converts input data to a list of FeatureAttribute instances.
    """

    _data: dict | None
    _items: list[FeatureAttribute] = field(init=False)

    def __post_init__(self) -> None:
        if self._data is not None:
            self._items = [
                FeatureAttribute(key, value) for key, value in self._data.items()
            ]
        else:
            self._items = []
        clear_temp_data(self)

    def __iter__(self) -> Iterator[FeatureAttribute]:
        yield from self._items

    @overload
    def __getitem__(self, idx: SupportsIndex) -> FeatureAttribute: ...

    @overload
    def __getitem__(self, idx: slice) -> list[FeatureAttribute]: ...

    def __getitem__(
        self, idx: SupportsIndex | slice
    ) -> FeatureAttribute | list[FeatureAttribute]:
        """Get a feature attribute by index or slice."""
        return self._items[idx]

    def __len__(self) -> int:
        return len(self._items)

    def __delitem__(self, idx):
        raise NotImplementedError("This method is not allowed in FeatureAttributes")

    def __setitem__(self, idx, value):
        raise NotImplementedError("This method is not allowed in FeatureAttributes")

    def __repr__(self) -> str:
        return repr(self._items)

    def set_attribute(self, field_: FeatureField, function: Callable) -> None:
        for item in self._items:
            if item.key == field_.name:
                item.value = function(item.value)
                break
        else:
            raise KeyError(f"Key '{field_.name}' not found in FeatureAttributes")  # noqa: TRY003

    def get_all_fields(self) -> list[str]:
        """
        Get all field names.

        Returns:
            List[str]: List of all field names.
        """
        return [item.key for item in self._items]

    def get_value(self, key: str) -> FeatureAttribute | None:
        """
        Get the value of a field.

        Parameters:
            key (str): The key of the field.

        Returns:
            Optional[FeatureAttribute]: The value of the field.
        """
        response = [item.value for item in self._items if item.key == key]
        if len(response) == 0:
            return None
        elif len(response) > 1:
            raise Exception(f"Duplicated key: {key}")  # noqa: TRY002, TRY003
        return response[0]


def parse_date(date_value: int) -> datetime | None:
    """
    Parse a date value.

    Parameters:
        date_value (int): The date value to parse.

    Returns:
        Optional[datetime]: The parsed date.
    """
    return datetime.utcfromtimestamp(date_value / 1000)



class ImmutableList:
    def __init__(self, items):
        self._items = tuple(items)

    def __getitem__(self, index):
        return self._items[index]

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return iter(self._items)

    def __contains__(self, item):
        return item in self._items

    def __repr__(self):
        return f"ImmutableList({self._items})"

    # Additional methods for your use case:
    def index(self, item):
        return self._items.index(item)

    def count(self, item):
        return self._items.count(item)

    def __eq__(self, other):
        if isinstance(other, ImmutableList):
            return self._items == other._items
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, ImmutableList):
            return self._items != other._items
        return NotImplemented

    def __add__(self, other):
        if isinstance(other, ImmutableList):
            return ImmutableList(list(self._items) + list(other._items))
        raise TypeError(  # noqa: TRY003
            f"unsupported operand type(s) for +: 'ImmutableList' and {type(other).__name__}"  # noqa: TRY003
        )  # noqa: TRY003

    def __radd__(self, other):
        if isinstance(other, (list | tuple)):
            return ImmutableList(list(other) + list(self._items))
        raise TypeError(  # noqa: TRY003
            f"unsupported operand type(s) for +: {type(other).__name__} and 'ImmutableList'"  # noqa: TRY003
        )  # noqa: TRY003

    def __mul__(self, n):
        return ImmutableList(list(self._items) * n)

    def __rmul__(self, n):
        return self.__mul__(n)



from arcGisFeatureCache.featureService import (
    ArcGisFeatureLayer,
    ArcGisFeatureService,
    ArcGisScraper,
    get_feature_service,
)

__version__ = "0.1.0a1"

__all__ = [
    "ArcGisFeatureService",
    "ArcGisFeatureLayer",
    "ArcGisScraper",
    "get_feature_service",
]



import asyncio

import numpy as np
from numpy.typing import NDArray
from shapely import STRtree  # type: ignore

from arcGisFeatureCache.arcGisScraper import ArcGisScraper
from arcGisFeatureCache.featureLayer import ArcGisFeatureLayer
from arcGisFeatureCache.models.feature import Feature


class ArcGisFeatureService:
    """A class representing an ArcGIS Feature Service.

    Attributes:
        name (str): The name of the feature service extracted from the URL.
        feature_service_layers (List[ArcGisFeatureLayer]): List of layers in the feature service.

    """

    def __init__(self, feature_service_url: str):
        self._base_url: str = feature_service_url
        self.name: str = feature_service_url.split(r"/")[-2]
        self.feature_service_layers: list[ArcGisFeatureLayer] = []
        self._tree: STRtree = STRtree([])
        self._tree_keys: NDArray = np.array([])

    @classmethod
    async def factory(cls, url) -> "ArcGisFeatureService":
        """
        Create an ArcGisFeatureService instance.

        Parameters:
            url: The URL of the ArcGIS Feature Service.

        Returns:
            An instance of ArcGisFeatureService.
        """
        self = ArcGisFeatureService(url)
        arc_gis_scraper = await ArcGisScraper.factory(url)
        layer_legenda = {
            f"{item['id']}": item["name"]
            for item in arc_gis_scraper.feature_service_info["layers"]
        }

        # TODO: get layer name here so we allso can use it when no feature returns (returns count=0)

        tasks = [
            ArcGisFeatureLayer.factory(item, layer_legenda)
            for item in arc_gis_scraper.feature_layers
        ]
        self.feature_service_layers = await asyncio.gather(*tasks)
        records = [
            {"geometry": item.geometry, "key": item.uuid}
            for layer in self.feature_service_layers
            for item in layer.features
        ]

        self._tree = STRtree([record["geometry"] for record in records])
        self._tree_keys = np.array([record["key"] for record in records])

        return self

    def get_all_features(self, feature_layer: list[str] | None = None) -> list[Feature]:
        """
        Get all features from the feature service.

        Parameters:
            feature_layer List of layer names to filter features (default: None).

        Returns:
            List of all features.
        """
        out_list = []
        for item in self.feature_service_layers:
            if feature_layer:
                out_list.extend(
                    [item2 for item2 in item.features if item2.dataset in feature_layer]
                )
            else:
                out_list.extend([item2 for item2 in item.features])
        return out_list

    def get_all_datasets(self) -> list[str]:
        """
        Retrieve a list of all datasets present..

        Returns:
            list[str]: A list containing the dataset names of all feature service layers.
        """
        return [_.dataset for _ in self.feature_service_layers]

    def get_layer_features(self, layers: list[str]) -> list[Feature]:
        """
        Get all features of a or some by dataset name(s).

        Parameters:
            layers: List of UUIDs to search for.

        Returns:
            List of features with matching dataset names.

        """
        out_list = []
        for item in self.feature_service_layers:
            out_list.extend([_ for _ in item.features if _.dataset in layers])
        return out_list

    # def get_features_in_box(
    #     self, bbox: box, layers: list[str] | None = None
    # ) -> list[Feature]:
    #     if layers is None:
    #         layers = []
    #     """
    #     Get features within a bounding box.
    #
    #     Parameters:
    #         bbox: The bounding box (Shapely box object).
    #         layers: layers to query
    #
    #     Returns:
    #         List of features within the bounding box.
    #
    #     """
    #     query = self._tree.query(bbox)
    #     keys = self._tree_keys.take(query).tolist()
    #     return self.get_layer_features(keys)

    # def get_features_matching_field_value(
    #     self,
    #     field: FeatureAttribute,
    #     value: Any,
    #     feature_layer: Optional[List[str]] = None,
    # ):
    #     # TODO: attribute query on layer
    #     out_list = []
    #     for item in self.feature_service_layers:
    #         pass
    #     return out_list


def get_feature_service(url: str) -> ArcGisFeatureService:
    return asyncio.run(ArcGisFeatureService.factory(url))



import asyncio
import uuid
from collections.abc import Callable, Iterator
from dataclasses import dataclass, field
from typing import SupportsIndex, overload

from shapely import LineString, Point, Polygon  # type: ignore

from arcGisFeatureCache.models.attributes import FeatureAttributes, FeatureField
from arcGisFeatureCache.models.geometry import (
    EsriGeometryTypeEnum,
    parse_line,
    parse_point,
    parse_polygon,
)
from arcGisFeatureCache.utils.helpers import clear_temp_data


@dataclass
class Feature:
    """A class representing a feature in an ArcGIS feature service.

    Generates a UUID for the feature and initializes its attributes.

    """

    _data: dict | None

    attributes: FeatureAttributes = field(init=False)
    geometry: Point | LineString | Polygon = field(init=False)
    uuid: str = field(init=False)
    dataset: str | None = field(init=False)
    measure_geometry: LineString | None = field(default=None)

    def __post_init__(self) -> None:
        self.uuid = str(uuid.uuid4())
        if self._data is not None and "attributes" in self._data.keys():
            self.attributes = FeatureAttributes(self._data["attributes"])

    def __repr__(self) -> str:
        return f"<Feature {self.attributes.__dict__}>"

    async def set_geometry(self, function: Callable) -> None:
        """
        Set the geometry of the feature.

        Parameters:
            function (Callable): The function to parse and set the geometry.
        """
        if self._data is not None:
            self.geometry = await function(self._data["geometry"])

    async def set_m_values(self) -> None:
        """
        Set M values for the feature's geometry.
        """
        if self._data is not None:
            geometry_data = self._data["geometry"]
        else:
            geometry_data = []

        if (
            "hasM" in geometry_data
            and geometry_data["hasM"]
            and len(geometry_data["paths"]) != 0
        ):
            if "hasZ" in geometry_data and geometry_data["hasZ"]:
                self.measure_geometry = LineString(
                    [[item[0], item[2], item[3]] for item in geometry_data["paths"][0]]
                )
            else:
                self.measure_geometry = LineString(
                    [[item[0], item[1], item[2]] for item in geometry_data["paths"][0]]
                )

    async def clean_data(self) -> None:
        """Asynchronously clean temporary data."""

        clear_temp_data(self)


@dataclass
class Features(list):
    """A class representing a collection of features.

    Converts input data to a list of Feature instances.
    """

    _data: dict | None
    _items: list[Feature] = field(init=False)

    def __post_init__(self) -> None:
        if self._data is not None:
            self._items = [Feature(_) for _ in self._data]
        clear_temp_data(self)

    def __iter__(self) -> Iterator[Feature]:
        yield from self._items

    @overload
    def __getitem__(self, idx: SupportsIndex) -> Feature: ...

    @overload
    def __getitem__(self, idx: slice) -> list[Feature]: ...

    def __getitem__(self, idx: SupportsIndex | slice) -> Feature | list[Feature]:
        """Get a feature by index or slice."""

        return self._items[idx]

    def __repr__(self) -> str:
        return str(self._items)

    def _set_value(self, field_list: list[FeatureField], function: Callable) -> None:
        """
        Set values for specific fields in the feature attributes.

        Parameters:
            field_list (List[FeatureField]): List of fields to set values for.
            function (Callable): The function to apply to the attribute values.
        """
        for feature in self._items:
            for field_item in field_list:
                feature.attributes.set_attribute(field_item, function)

    async def set_geometry(self, geometry_type: EsriGeometryTypeEnum) -> None:
        """
        Set the geometry for all features.

        Parameters:
            geometry_type (EsriGeometryTypeEnum): The type of geometry to parse.
        """
        match geometry_type:
            case EsriGeometryTypeEnum.esriGeometryPoint:
                function = parse_point
            case EsriGeometryTypeEnum.esriGeometryPolyline:
                function = parse_line
            case EsriGeometryTypeEnum.esriGeometryPolygon:
                function = parse_polygon
            case _:
                raise NotImplementedError

        tasks = [feature.set_geometry(function) for feature in self._items]
        await asyncio.gather(*tasks)

    async def set_measure(self) -> None:
        """
        Set M values for all features.
        """
        tasks = [feature.set_m_values() for feature in self._items]
        await asyncio.gather(*tasks)

    async def clean_data(self) -> None:
        """
        Clean temporary data for all features.
        """
        tasks = [feature.clean_data() for feature in self._items]
        await asyncio.gather(*tasks)



from dataclasses import dataclass, field
from enum import Enum

from shapely import LineString, Point, Polygon  # type: ignore

from arcGisFeatureCache.utils.helpers import clear_temp_data


class EsriGeometryTypeEnum(Enum):
    """
    Enumeration representing Esri geometry types.

    Attributes:
        esriGeometryPoint: esri Point geometry.
        esriGeometryMultipoint: esri Multipoint geometry.
        esriGeometryPolyline: esri Polyline geometry.
        esriGeometryPolygon: esri Polygon geometry.
        esriGeometryEnvelope: esri Envelope geometry.
    """

    esriGeometryPoint = "esriGeometryPoint"
    esriGeometryMultipoint = "esriGeometryMultipoint"
    esriGeometryPolyline = "esriGeometryPolyline"
    esriGeometryPolygon = "esriGeometryPolygon"
    esriGeometryEnvelope = "esriGeometryEnvelope"


@dataclass
class FeatureSpatialReference:
    """
    Dataclass representing the spatial reference of a feature.

    Attributes:
        wkid (str): Well-known ID (WKID) of the spatial reference.
        latest_wkid (str): Latest well-known ID (WKID) of the spatial reference.
    """

    _data: dict | None
    wkid: str = field(init=False)
    latest_wkid: str = field(init=False)

    def __post_init__(self):
        if not hasattr(self, "wkid"):
            self.wkid = self._data["wkid"]

        if not hasattr(self, "latestWkid"):
            self.latest_wkid = self._data["latestWkid"]

        clear_temp_data(self)


async def parse_point(geometry_value: dict) -> Point:
    """
    Asynchronously parse a point geometry.

    Parameters:
        geometry_value (Dict): Dictionary containing point geometry data.

    Returns:
        Point: Shapely Point object representing the parsed point geometry.
    """
    # TODO: implement 3d
    if geometry_value == "":
        return Point()
    return Point(geometry_value["x"], geometry_value["y"])


async def parse_line(geometry_value: dict) -> LineString:
    """
    Asynchronously parse a line geometry.

    Parameters:
        geometry_value (Dict): Dictionary containing line geometry data.

    Returns:
        LineString: Shapely LineString object representing the parsed line geometry.
    """
    if geometry_value == {} or (
        "paths" in geometry_value and len(geometry_value["paths"]) == 0
    ):
        return LineString()
    elif "hasZ" in geometry_value.keys() and geometry_value["hasZ"]:
        return LineString(
            [[item[0], item[1], item[2]] for item in geometry_value["paths"][0]]
        )
    else:
        return LineString(geometry_value["paths"][0])


async def parse_polygon(geometry_value: dict) -> Polygon:
    """
    Asynchronously parse a polygon geometry.

    Parameters:
        geometry_value (Dict): Dictionary containing polygon geometry data.

    Returns:
        Polygon: Shapely Polygon object representing the parsed polygon geometry.
    """
    # TODO: implement 3d and inner outer stuff
    if geometry_value == {}:
        return Polygon()
    return Polygon(geometry_value["rings"][0])



from dataclasses import dataclass, field

from arcGisFeatureCache.models.attributes import FeatureFields, parse_date
from arcGisFeatureCache.models.feature import Features
from arcGisFeatureCache.models.geometry import (
    EsriGeometryTypeEnum,
    FeatureSpatialReference,
)
from arcGisFeatureCache.utils.helpers import clear_temp_data
from arcGisFeatureCache.utils.log import logger


@dataclass
class ArcGisFeatureLayer:
    """A class representing a layer in an ArcGIS Feature Service.

    Sets attributes based on input data.
    """

    _data: dict

    object_id_field_name: str = field(init=False, default="")
    global_id_field_name: str = field(init=False, default="")
    geometry_type: EsriGeometryTypeEnum = field(init=False)
    spatial_reference: FeatureSpatialReference = field(init=False)
    fields: FeatureFields = field(init=False)
    features: Features = field(init=False)
    dataset: str = field(init=False, default="")
    has_m: bool = False

    def __post_init__(self):
        self.object_id_field_name = self._data["objectIdFieldName"]
        self.global_id_field_name = self._data["globalIdFieldName"]
        self.geometry_type = EsriGeometryTypeEnum[self._data["geometryType"]]
        self.spatial_reference = FeatureSpatialReference(self._data["spatialReference"])
        self.fields = FeatureFields(self._data["fields"])
        self.features = Features(self._data["features"])
        self.dataset = self._data["dataset"].split(r"/")[-1]

        clear_temp_data(self)

    async def _set_dataset_name(self, layer_legenda: dict) -> None:
        """
        sets the layer name based on a legend.

        Parameters:
            layer_legenda: A dictionary containing the legend for layer names.
        """
        self.dataset = layer_legenda[self.dataset]
        for item in self.features:
            item.dataset = self.dataset

    async def _process_dates(self) -> None:
        """
        processes date fields in features.
        """
        self.features._set_value(self.fields.get_date_fields(), parse_date)

    async def _process_geometry(self) -> None:
        """
        Processes geometry in features.
        """
        await self.features.set_geometry(self.geometry_type)
        await self.features.set_measure()
        await self.features.clean_data()

    @classmethod
    async def factory(cls, _data: dict, layer_legenda: dict) -> "ArcGisFeatureLayer":
        """
        Creates an ArcGisFeatureServiceLayer instance.

        Parameters:
            _data: Data for creating the layer.
            layer_legenda: Legend for layer names.

        Returns:
            An instance of ArcGisFeatureServiceLayer.
        """
        self = ArcGisFeatureLayer(_data)
        dataset_name = layer_legenda[_data["dataset"].split("/")[-1]]

        logger.debug(f"Processing new Feature Layer: {dataset_name}")
        await self._set_dataset_name(layer_legenda)
        await self._process_dates()
        await self._process_geometry()
        if len(self.features) != 0:
            self.has_m = False
        else:
            self.has_m = (
                True if self.features[0].measure_geometry is not None else False
            )

        logger.debug(f"Done, created dataset: {dataset_name}")

        return self

    def get_all_features(self) -> Features:
        """
        Get all features from the dataset.

        Returns:
            All features in the layer.
        """
        return self.features



import asyncio

import pytest

from arcGisFeatureCache import ArcGisFeatureService

feature_service_instance_km = asyncio.run(
    ArcGisFeatureService.factory(
        "https://mapservices.prorail.nl/arcgis/rest/services/Referentiesysteem_004/FeatureServer"
    )
)


@pytest.mark.parametrize(
    "feature_service_instance,feature_layer,layers_count,layer_name",
    [
        (
            feature_service_instance_km,
            "Referentiesysteem_004",
            9,
            "Koppelpunt kilometerlint",
        ),
    ],
)
@pytest.mark.asyncio
async def test_feature_service_layers(
    feature_service_instance: ArcGisFeatureService,
    feature_layer: str,
    layers_count: int,
    layer_name: str,
):
    assert len(feature_service_instance.feature_service_layers) == layers_count
    assert feature_service_instance.name == feature_layer

    layer_features = feature_service_instance.get_layer_features([layer_name])
    assert layer_features[0].dataset == layer_name
    assert layer_features[-1].dataset == layer_name


@pytest.mark.parametrize(
    "feature_service_instance",
    [
        feature_service_instance_km,
    ],
)
def test_feature_service(feature_service_instance: ArcGisFeatureService):
    all_layers = [_.dataset for _ in feature_service_instance.feature_service_layers]
    for layer_name in all_layers:
        first_item = feature_service_instance.get_all_features([layer_name])[0]
        last_item = feature_service_instance.get_all_features([layer_name])[-1]
        assert first_item.dataset == layer_name
        assert last_item.dataset == layer_name
        for feature in [
            _
            for _ in feature_service_instance.feature_service_layers
            if _.dataset == layer_name
        ][0].features:
            assert feature.dataset == layer_name
            # TODO: check if line point polygon or else
            assert feature.geometry is not None
            all_fields = feature.attributes.get_all_fields()
            for field in all_fields:
                assert feature.attributes.get_value(field) is not any
